"""
Decoder for RuuviTag Data Format 5 data.
Based on https://github.com/Bluetooth-Devices/ruuvitag-ble/blob/0e99249/src/ruuvitag_ble/df5_decoder.py (MIT Licensed)  # noqa
Which was based on https://github.com/ttu/ruuvitag-sensor/blob/23e6555/ruuvitag_sensor/decoder.py (MIT Licensed)  # noqa
"""
from __future__ import annotations

import math
import struct
import typing as ty


class DataFormat5Decoder:
    def __init__(self, raw_data: bytes) -> None:
        if len(raw_data) < 24:
            raise ValueError(
                "Data must be at least 24 bytes long for data format 5",
            )
        self.data = struct.unpack(">BhHHhhhHBH6B", raw_data)

    @property
    def temperature_celsius(self) -> ty.Optional[float]:
        if self.data[1] == -32768:
            return None
        return round(self.data[1] / 200.0, 2)

    @property
    def humidity_percentage(self) -> ty.Optional[float]:
        if self.data[2] == 65535:
            return None
        return round(self.data[2] / 400, 2)

    @property
    def pressure_hpa(self) -> ty.Optional[float]:
        if self.data[3] == 0xFFFF:
            return None

        return round((self.data[3] + 50000) / 100, 2)

    @property
    def acceleration_vector_mg(self) \
            -> ty.Union[
                ty.Sequence[int, int, int],
                ty.Sequence[None, None, None],
            ]:
        ax = self.data[4]
        ay = self.data[5]
        az = self.data[6]
        if ax == -32768 or ay == -32768 or az == -32768:
            return None, None, None

        return ax, ay, az

    @property
    def acceleration_total_mg(self) -> ty.Optional[float]:
        ax, ay, az = self.acceleration_vector_mg
        if ax is None or ay is None or az is None:
            return None
        return math.sqrt(ax * ax + ay * ay + az * az)

    @property
    def battery_voltage_mv(self) -> ty.Optional[int]:
        voltage = self.data[7] >> 5
        if voltage == 0b11111111111:
            return None

        return voltage + 1600

    @property
    def tx_power_dbm(self) -> ty.Optional[int]:
        tx_power = self.data[7] & 0x001F
        if tx_power == 0b11111:
            return None

        return -40 + (tx_power * 2)

    @property
    def movement_counter(self) -> int:
        return self.data[8]

    @property
    def measurement_sequence_number(self) -> int:
        return self.data[9]

    @property
    def mac(self) -> str:
        return ":".join(f"{x:02X}" for x in self.data[10:])



import logging
from dataclasses import dataclass

from bleak.backends.device import BLEDevice

from ..devices.base import SENSOR_DOMAIN, Sensor
from ..protocols.ruuvi import DataFormat5Decoder
from ..utils import cr2477_voltage_to_percent, format_binary

_LOGGER = logging.getLogger(__name__)


@dataclass
class SensorState:
    battery: int = 0
    temperature: float = 0
    humidity: float = 0
    pressure: float = 0
    movement_counter: int = 0


class RuuviTag(Sensor):
    NAME = 'ruuvitag'
    MANUFACTURER = 'Ruuvi'
    SENSOR_CLASS = SensorState
    SUPPORT_PASSIVE = True
    SUPPORT_ACTIVE = False
    # send data only if temperature or humidity is set
    REQUIRED_VALUES = ('temperature', 'humidity', 'pressure')

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'humidity',
                    'device_class': 'humidity',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'pressure',
                    'device_class': 'atmospheric_pressure',
                    'unit_of_measurement': 'hPa',
                },
                {
                    'name': 'movement_counter',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }

    def handle_advert(self, scanned_device: BLEDevice, adv_data):
        raw_data = adv_data.manufacturer_data[0x0499]

        data_format = raw_data[0]
        if data_format != 0x05:
            _LOGGER.debug("Data format not supported: %s", raw_data[0])
            return

        if raw_data:
            decoder = DataFormat5Decoder(bytes(raw_data))
            self._state = self.SENSOR_CLASS(
                temperature=decoder.temperature_celsius,
                humidity=decoder.humidity_percentage,
                pressure=decoder.pressure_hpa,
                movement_counter=decoder.movement_counter,
                battery=int(cr2477_voltage_to_percent(
                    decoder.battery_voltage_mv,
                ))
            )

            _LOGGER.debug(
                f'Advert received for {self}, {format_binary(raw_data)}, '
                f'current state: {self._state}',
            )


class RuuviTagPro2in1(RuuviTag):
    NAME = 'ruuvitag_pro_2in1'

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'movement_counter',
                    'device_class': 'count',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }


class RuuviTagPro3in1(RuuviTag):
    NAME = 'ruuvitag_pro_3in1'

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'humidity',
                    'device_class': 'humidity',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'movement_counter',
                    'device_class': 'count',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }



import asyncio as aio
import json
import logging
import typing as ty
from uuid import getnode

import aio_mqtt
from bleak.backends.device import BLEDevice

from .compat import get_scanner
from .devices.base import ConnectionMode, Device, done_callback
from .exceptions import (ListOfConnectionErrors, handle_ble_exceptions,
                         restart_bluetooth)
from .manager import DeviceManager
from .tasks import handle_returned_tasks, run_tasks_and_cancel_on_first_return

try:
    from bleak import AdvertisementData
except ImportError:
    AdvertisementData = ty.Any

_LOGGER = logging.getLogger(__name__)

BRIDGE_STATE_TOPIC = 'state'


class Ble2Mqtt:
    TOPIC_ROOT = 'ble2mqtt'
    BRIDGE_TOPIC = 'bridge'

    def __init__(
            self,
            ssl,
            host: str,
            port: int = None,
            user: ty.Optional[str] = None,
            password: ty.Optional[str] = None,
            reconnection_interval: int = 10,
            loop: ty.Optional[aio.AbstractEventLoop] = None,
            *,
            hci_adapter: str,
            base_topic,
            mqtt_config_prefix,
            legacy_color_mode,
    ) -> None:
        self._hci_adapter = hci_adapter
        self._mqtt_host = host
        self._mqtt_port = port
        self._mqtt_user = user
        self._mqtt_password = password
        self._ssl = ssl
        self._base_topic = base_topic
        self._mqtt_config_prefix = mqtt_config_prefix
        self._legacy_color_mode = legacy_color_mode

        self._reconnection_interval = reconnection_interval
        self._loop = loop or aio.get_event_loop()

        self._mqtt_client = aio_mqtt.Client(
            client_id_prefix=f'{base_topic}_',
            loop=self._loop,
        )

        self._device_managers: ty.Dict[Device, DeviceManager] = {}

        self.availability_topic = '/'.join((
            self._base_topic,
            self.BRIDGE_TOPIC,
            BRIDGE_STATE_TOPIC,
        ))

        self.device_registry: ty.List[Device] = []

    async def start(self):
        result = await run_tasks_and_cancel_on_first_return(
            self._loop.create_task(self._connect_mqtt_forever()),
            self._loop.create_task(self._handle_messages()),
        )
        for t in result:
            await t

    async def close(self) -> None:
        for device, manager in self._device_managers.items():
            await manager.close()

        if self._mqtt_client.is_connected:
            try:
                await self._mqtt_client.disconnect()
            except aio.CancelledError:
                raise
            except aio_mqtt.ConnectionClosedError:
                pass
            except Exception as e:
                _LOGGER.warning(f'Error on MQTT disconnecting: {repr(e)}')

    def register(self, device_class: ty.Type[Device], *args, **kwargs):
        device = device_class(*args, **kwargs)
        if not device:
            return
        if not device.is_passive and not device.SUPPORT_ACTIVE:
            raise NotImplementedError(
                f'Device {device.dev_id} doesn\'t support active mode',
            )
        assert device.is_passive or device.ACTIVE_CONNECTION_MODE in (
            ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT,
            ConnectionMode.ACTIVE_KEEP_CONNECTION,
            ConnectionMode.ON_DEMAND_CONNECTION,
        )
        self.device_registry.append(device)

    @property
    def subscribed_topics(self):
        return [
            '/'.join((self._base_topic, topic))
            for device in self.device_registry
            for topic in device.subscribed_topics
        ]

    async def _handle_messages(self) -> None:
        async for message in self._mqtt_client.delivered_messages(
            f'{self._base_topic}/#',
        ):
            _LOGGER.debug(message)
            while True:
                if message.topic_name not in self.subscribed_topics:
                    await aio.sleep(0)
                    continue

                prefix = f'{self._base_topic}/'
                if message.topic_name.startswith(prefix):
                    topic_wo_prefix = message.topic_name[len(prefix):]
                else:
                    topic_wo_prefix = prefix
                for _device in self.device_registry:
                    if topic_wo_prefix in _device.subscribed_topics:
                        device = _device
                        break
                else:
                    raise NotImplementedError('Unknown topic')
                await aio.sleep(0)
                if not device.client.is_connected:
                    _LOGGER.warning(
                        f'Received topic {topic_wo_prefix} '
                        f'with {message.payload} '
                        f' but {device.client} is offline',
                    )
                    await aio.sleep(5)
                    continue

                try:
                    value = json.loads(message.payload)
                except ValueError:
                    value = message.payload.decode()

                await device.add_incoming_message(topic_wo_prefix, value)
                break

            await aio.sleep(1)

    async def stop_device_manage_tasks(self):
        for manager in self._device_managers.values():
            try:
                await manager.close()
            except aio.CancelledError:
                raise
            except Exception:
                _LOGGER.exception(
                    f'Problem on closing dev manager {manager.device}')

    def device_detection_callback(self, device: BLEDevice,
                                  advertisement_data: AdvertisementData):
        for reg_device in self.device_registry:
            if reg_device.mac.lower() == device.address.lower():
                if hasattr(advertisement_data, 'rssi'):
                    rssi = advertisement_data.rssi
                else:
                    rssi = device.rssi
                if rssi:
                    # update rssi for all devices if available
                    reg_device.rssi = rssi

                if reg_device in self._device_managers:
                    self._device_managers[reg_device].set_scanned_device(device)

                if reg_device.is_passive:
                    if device.name:
                        reg_device._model = device.name
                    reg_device.handle_advert(device, advertisement_data)
                else:
                    _LOGGER.debug(
                        f'active device seen: {reg_device} '
                        f'{advertisement_data}',
                    )
                    reg_device.set_advertisement_seen()

    async def scan_devices_task(self):
        empty_scans = 0
        while True:
            # 10 empty scans in a row means that bluetooth restart is required
            if empty_scans >= 10:
                empty_scans = 0
                await restart_bluetooth(self._hci_adapter)

            try:
                async with handle_ble_exceptions(self._hci_adapter):
                    scanner = get_scanner(
                        self._hci_adapter,
                        self.device_detection_callback,
                    )
                    try:
                        await aio.wait_for(scanner.start(), 10)
                    except aio.TimeoutError:
                        _LOGGER.error('Scanner start failed with timeout')
                    await aio.sleep(3)
                    devices = scanner.discovered_devices
                    await scanner.stop()
                    if not devices:
                        empty_scans += 1
                    else:
                        empty_scans = 0
                    _LOGGER.debug(f'found {len(devices)} devices: {devices}')
            except KeyboardInterrupt:
                raise
            except aio.IncompleteReadError:
                raise
            except ListOfConnectionErrors as e:
                _LOGGER.exception(e)
                empty_scans += 1
            await aio.sleep(1)

    async def _run_device_tasks(self, mqtt_connection_fut: aio.Future) -> None:
        for dev in self.device_registry:
            self._device_managers[dev] = \
                DeviceManager(
                    dev,
                    hci_adapter=self._hci_adapter,
                    mqtt_client=self._mqtt_client,
                    base_topic=self._base_topic,
                    config_prefix=self._mqtt_config_prefix,
                    global_availability_topic=self.availability_topic,
                    legacy_color_mode=self._legacy_color_mode,
                )
        _LOGGER.debug("Wait for network interruptions...")

        device_tasks = [
            manager.run_task()
            for manager in self._device_managers.values()
        ]
        scan_task = self._loop.create_task(self.scan_devices_task())
        scan_task.add_done_callback(done_callback)
        device_tasks.append(scan_task)

        futs = [
            mqtt_connection_fut,
            *device_tasks,
        ]

        finished = await run_tasks_and_cancel_on_first_return(
            *futs,
            ignore_futures=[mqtt_connection_fut],
        )

        finished_managers = []
        for d, m in self._device_managers.items():
            if m.manage_task not in finished:
                await m.close()
            else:
                finished_managers.append(m)

        for m in finished_managers:
            await m.close()

        # when mqtt server disconnects, multiple tasks can raise
        # exceptions. We must fetch all of them
        finished = [t for t in futs if t.done() and not t.cancelled()]
        await handle_returned_tasks(*finished)

    async def _connect_mqtt_forever(self) -> None:
        dev_id = hex(getnode())
        while True:
            try:
                mqtt_connection = await aio.wait_for(self._mqtt_client.connect(
                    host=self._mqtt_host,
                    port=self._mqtt_port,
                    username=self._mqtt_user,
                    password=self._mqtt_password,
                    ssl=self._ssl,
                    client_id=f'ble2mqtt_{dev_id}',
                    will_message=aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='offline',
                        qos=aio_mqtt.QOSLevel.QOS_1,
                        retain=True,
                    ),
                ), timeout=self._reconnection_interval)
                _LOGGER.info(f'Connected to {self._mqtt_host}')
                await self._mqtt_client.publish(
                    aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='online',
                        qos=aio_mqtt.QOSLevel.QOS_1,
                        retain=True,
                    ),
                )
                await self._run_device_tasks(mqtt_connection.disconnect_reason)
            except aio.TimeoutError:
                logging.warning('Cannot connect to MQTT broker')
            except (aio.CancelledError, KeyboardInterrupt):
                if self._mqtt_client.is_connected():
                    await self._mqtt_client.publish(
                        aio_mqtt.PublishableMessage(
                            topic_name=self.availability_topic,
                            payload='offline',
                            qos=aio_mqtt.QOSLevel.QOS_0,
                            retain=True,
                        ),
                    )
                raise
            except Exception:
                _LOGGER.exception(
                    "Connection lost. Will retry in %d seconds.",
                    self._reconnection_interval,
                )
                try:
                    await self.stop_device_manage_tasks()
                except aio.CancelledError:
                    raise
                except Exception:
                    _LOGGER.exception('Exception in _connect_forever()')
                try:
                    await self._mqtt_client.disconnect()
                except aio.CancelledError:
                    raise
                except Exception:
                    _LOGGER.error('Disconnect from MQTT broker error')
                await aio.sleep(self._reconnection_interval)



import logging

from bleak.backends.device import BLEDevice

from ..devices.base import HumidityTemperatureSensor
from ..protocols.govee import GoveeDecoder
from ..utils import format_binary

_LOGGER = logging.getLogger(__name__)

def valid_data_length(raw_data):
    valid_lengths = [6, 7]
    return len(raw_data) in valid_lengths


class GoveeTemperature(HumidityTemperatureSensor):
    NAME = 'govee_ht'
    MANUFACTURER = 'Govee'
    SUPPORT_PASSIVE = True
    SUPPORT_ACTIVE = False

    def handle_advert(self, scanned_device: BLEDevice, adv_data):
        raw_data = adv_data.manufacturer_data.get(0xec88)
        if not raw_data:
            _LOGGER.debug(
                'Temperature data not found; got '
                f'{repr(adv_data.manufacturer_data)}',
            )
            return

        if not valid_data_length(raw_data):
            _LOGGER.debug(
                'Unexpected raw data length '
                f'{len(raw_data)} ({repr(raw_data)})',
            )

        decoder = GoveeDecoder(bytes(raw_data))
        self._state = self.SENSOR_CLASS(
            temperature=decoder.temperature_celsius,
            humidity=decoder.humidity_percentage,
            battery=decoder.battery_percentage,
        )

        _LOGGER.debug(
            f'Advert received for {self}, {format_binary(raw_data)}, '
            f'current state: {self._state}',
        )



import asyncio as aio
import logging
import struct
import uuid
from dataclasses import dataclass

from .base import SENSOR_DOMAIN, ConnectionMode, Sensor
from .uuids import BATTERY

_LOGGER = logging.getLogger(__name__)

FLOWER_SENSOR_CHAR = uuid.UUID('55482920-eacb-11e3-918a-0002a5d5c51b')

TEMPERATURE_VALUES = [68.8, 49.8, 24.3, 6.4, 1.0, -5.5, -20.5, -41.0]
TEMPERATURE_READINGS = [1035, 909, 668, 424, 368, 273, 159, 0]
MOISTURE_VALUES = [60.0, 58.0, 54.0, 22.0, 2.0, 0.0]
MOISTURE_READINGS = [1254, 1249, 1202, 1104, 944, 900]
LIGHT_VALUES = [
    175300.0, 45400.0, 32100.0, 20300.0, 14760.0, 7600.0, 1200.0, 444.0,
    29.0, 17.0, 0.0,
]
LIGHT_READINGS = [911, 764, 741, 706, 645, 545, 196, 117, 24, 17, 0]


def _interpolate(raw_value, values, raw_values):
    index = 0
    if raw_value > raw_values[0]:
        index = 0
    elif raw_value < raw_values[-2]:
        index = len(raw_values) - 2
    else:
        while raw_value < raw_values[index + 1]:
            index += 1

    delta_value = values[index] - values[index + 1]
    delta_raw = raw_values[index] - raw_values[index + 1]
    return (
        (raw_value - raw_values[index + 1]) * delta_value / delta_raw +
        values[index + 1]
    )


def calculate_temperature(raw_value):
    return _interpolate(raw_value, TEMPERATURE_VALUES, TEMPERATURE_READINGS)


def calculate_moisture(raw_value):
    moisture = _interpolate(raw_value, MOISTURE_VALUES, MOISTURE_READINGS)

    if moisture > 100.0:
        moisture = 100.0
    if moisture < 0.0:
        moisture = 0.0

    return moisture


def calculate_illuminance(raw_value):
    return _interpolate(raw_value, LIGHT_VALUES, LIGHT_READINGS)


@dataclass
class SensorState:
    temperature: float
    moisture: float
    illuminance: int
    battery: int = 0

    @classmethod
    def from_data(cls, data: bytes, battery: bytes):
        temp_raw, moisture_raw, illuminance_raw = struct.unpack('<HxxHH', data)
        return cls(
            temperature=round(calculate_temperature(temp_raw), 2),
            moisture=round(calculate_moisture(moisture_raw), 2),
            illuminance=int(calculate_illuminance(illuminance_raw)),
            battery=int.from_bytes(battery, byteorder='little'),
        )


class FlowerMonitorMCLH09(Sensor):
    NAME = 'mclh09'
    MANUFACTURER = 'LifeControl'
    DATA_CHAR = FLOWER_SENSOR_CHAR
    BATTERY_CHAR = BATTERY
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT
    DEFAULT_RECONNECTION_SLEEP_INTERVAL = 300
    READ_DATA_IN_ACTIVE_LOOP = True

    def __init__(self, mac, *args, **kwargs) -> None:
        super().__init__(mac, *args, **kwargs)
        self.RECONNECTION_SLEEP_INTERVAL = int(
            kwargs.get('interval', self.DEFAULT_RECONNECTION_SLEEP_INTERVAL)
        )

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'moisture',
                    'device_class': 'moisture',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'illuminance',
                    'device_class': 'illuminance',
                    'unit_of_measurement': 'lx',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }

    async def read_state(self):
        battery = await self._read_with_timeout(self.BATTERY_CHAR)
        data = await self._read_with_timeout(self.DATA_CHAR)
        for _ in range(5):
            try:
                self._state = SensorState.from_data(data, battery)
            except ValueError as e:
                _LOGGER.warning(f'{self} {repr(e)}')
                await aio.sleep(1)
            else:
                break

    async def do_active_loop(self, publish_topic):
        try:
            await aio.wait_for(self.read_state(), 5)
            await self._notify_state(publish_topic)
        except (aio.CancelledError, KeyboardInterrupt):
            raise
        except Exception:
            _LOGGER.exception(f'{self} problem with reading values')



"""
Decoder for Govee temperature/humidity sensors.
Based on https://github.com/wcbonner/GoveeBTTempLogger/blob/master/goveebttemplogger.cpp (MIT License)  # noqa: E501
"""
from __future__ import annotations
from enum import Enum

import struct
import typing as ty

class PartNumber(Enum):
    H5074 = 0
    H5075 = 1


def get_intermediate_temp_h5075(unpacked_bytes):
    """
    Decode the data and separate out the sign from the temperature. Note that
    the humidity and temperature are both encoded into the 24 bit returned
    value.
    """
    raw = int.from_bytes(unpacked_bytes, byteorder="big")
    is_negative = (raw & 0x800000) > 0
    temp = raw & 0x7FFFF

    return (temp, is_negative)


class GoveeDecoder:
    def __init__(self, raw_data: bytes) -> None:
        if len(raw_data) == 7:
            self.data = struct.unpack("<xhhBx", raw_data)
            self.part_number = PartNumber.H5074

        elif len(raw_data) == 6:
            self.data = struct.unpack("<x3sBx", raw_data)
            self.part_number = PartNumber.H5075

        else:
            raise ValueError("Govee data must be 6 or 7 bytes long")

    @property
    def temperature_celsius(self) -> ty.Optional[float]:
        if self.part_number == PartNumber.H5074:
            if self.data[0] == -32768:
                return None
            return round(self.data[0] / 100.0, 2)

        elif self.part_number == PartNumber.H5075:
            temp, is_negative = get_intermediate_temp_h5075(self.data[0])

            # This is temp/1000/10
            #see: https://github.com/wcbonner/GoveeBTTempLogger/issues/49
            temp /= 10000.0

            if is_negative:
                temp *= -1.0

            return temp

        else:
            return None

    @property
    def humidity_percentage(self) -> ty.Optional[float]:
        if self.part_number == PartNumber.H5074:
            if self.data[0] == -32768:
                return None
            return round(self.data[1] / 100.0, 2)

        elif self.part_number == PartNumber.H5075:
            temp, _ = get_intermediate_temp_h5075(self.data[0])
            humidity = (temp % 1000.0) / 10.0
            return humidity

        else:
            return None

    @property
    def battery_percentage(self) -> int:
        if self.part_number == PartNumber.H5074:
            return self.data[2]

        elif self.part_number == PartNumber.H5075:
            return self.data[1]

        else:
            return None



import logging

from bleak import AdvertisementData
from bleak.backends.device import BLEDevice

from ..devices.base import HumidityTemperatureSensor
from ..utils import format_binary

_LOGGER = logging.getLogger(__name__)

# Decoding was inspired by github.com/Trundle/meter-reader, github.com/Ernst79/bleparser & github.com/sblibs/pySwitchbot
# https://github.com/Trundle/meter-reader/blob/448a61dea487303817e902e7991708927a530847/src/meterreader_models/src/lib.rs#L98C12-L98C21
# https://github.com/Ernst79/bleparser/blob/6845901e04e164376a3f023c7a4a1f65db651cc7/package/bleparser/switchbot.py#L13
# https://github.com/sblibs/pySwitchbot/blob/3014972af637957f9dc29b2c4829faaec73f8526/switchbot/adv_parser.py
class SwitchBotHumidityTemperature(HumidityTemperatureSensor):
    MAC_TYPE = 'random'
    NAME = 'switchbot_ht'
    MANUFACTURER = 'SwitchBot'
    SUPPORT_PASSIVE = True
    SUPPORT_ACTIVE = False

    def handle_advert(self, scanned_device: BLEDevice, adv_data: AdvertisementData):
        data = adv_data.service_data.get("0000fd3d-0000-1000-8000-00805f9b34fb")
        if data is None:
            data = adv_data.service_data.get("00000d00-0000-1000-8000-00805f9b34fb")
        if not data:
            _LOGGER.debug(f'Service data not found for {self} got: {repr(adv_data)}')
            return

        if len(data) != 6:
            _LOGGER.error(f'Unexpected data length for {self} got: {format_binary(data)}')
            return

        if data[0] not in [84, 105]: # aka [0x54, 0x69] aka ["Meter", "Meter Plus"]
            _LOGGER.error(f'Unsupported model for {self} got: {format_binary(data)}')
            return

        temperature = float(data[4] & 127) + float((data[3] & 15) / 10.0)
        if data[4] & 128 == 0:
            temperature = -1 * temperature

        humidity = data[5] & 127
        battery = data[2] & 127

        self._state = self.SENSOR_CLASS(
            temperature=temperature,
            humidity=humidity,
            battery=battery,
        )



try:
    from importlib import metadata
except ImportError:
    metadata = None
import sys

import bleak


def get_loop_param(loop):
    if sys.version_info >= (3, 8):
        return {}
    return {'loop': loop}


def get_bleak_version():
    # returns None if version info is messed up
    if not metadata:
        return None
    return metadata.version('bleak')


bleak_version = get_bleak_version()


def get_scanner(hci_adapter: str, detection_callback) -> bleak.BleakScanner:
    if bleak_version and bleak_version < '0.18':
        scanner = bleak.BleakScanner(adapter=hci_adapter)
        scanner.register_detection_callback(detection_callback)
    else:
        scanner = bleak.BleakScanner(
            adapter=hci_adapter,
            detection_callback=detection_callback,
        )

    return scanner



import asyncio as aio
import logging
import typing as ty
import uuid
from dataclasses import asdict, dataclass
from enum import Enum

from ..compat import get_loop_param
from ..utils import format_binary
from .base import (BUTTON_DOMAIN, SELECT_DOMAIN, SENSOR_DOMAIN, ConnectionMode,
                   Sensor)
from .uuids import DEVICE_NAME

_LOGGER = logging.getLogger(__name__)


VOLTAGE_CHAR = uuid.UUID('0000ffd1-0000-1000-8000-00805f9b34fb')
STATE_CHAR = uuid.UUID('0000ffd2-0000-1000-8000-00805f9b34fb')
TOTAL_USE_CHAR = uuid.UUID('0000ffd3-0000-1000-8000-00805f9b34fb')
VERSION_CHAR = uuid.UUID('0000ffd4-0000-1000-8000-00805f9b34fb')
ERROR_CHAR = uuid.UUID('0000ffd6-0000-1000-8000-00805f9b34fb')

STATE_ENTITY = 'state'
VOLTAGE_ENTITY = 'voltage'
LOW_SPEED_POWER = 'low_speed_power'
RESET_HEPA = 'reset_hepa'


class CleanerStateEnum(Enum):
    UNKNOWN = 'unknown'
    WORKING = 'working'
    CLOSED = 'closed'
    STANDBY = 'standby'
    CHARGING = 'charging'


class SpeedOption(Enum):
    Low = 'Low'
    Medium = 'Medium'
    High = 'High'


class SpeedLvPower(Enum):
    _60W = 2550
    _80W = 3300
    _100W = 4200
    _110W = 4600
    _130W = 4950
    _150W = 5600
    _180W = 6450
    _200W = 7250
    _280W = 9500
    _300W = 10400
    _400W = 14000

    def to_option(self):
        return self.name[1:]

    @classmethod
    def from_option(cls, option: str):
        return cls[f'_{option}']

    @classmethod
    def _missing_(cls, value: int):
        # find closest to the value
        min_dist = None
        closest = None
        for item in cls.__members__.values():
            dist = abs(value - item.value)
            if min_dist is None or min_dist > dist:
                min_dist = dist
                closest = item
        return closest


class ErrorCode(Enum):
    OK = 0x51
    DustUnInstall = 0x54
    BrushBlocked = 0x52
    BrushDischarge = 0x53
    MotorAbnormal = 0x3857
    DustFull = 0x55
    TemperatureHigh = 0x56
    MIFDateout = 0xf9
    Unknown = 0x9999

    @classmethod
    def _missing_(cls, value):
        return cls.Unknown


@dataclass
class CleanerState:
    battery: int
    voltage: int
    total_use_seconds: int
    numeric_display_type: int
    total_use_seconds: int
    total_clean_seconds: int
    hepa_used_seconds: int
    speed_level: SpeedOption
    low_speed_lv_power: SpeedLvPower
    numeric_display_type: int
    state: CleanerStateEnum
    dust_abnormal_value: ErrorCode
    brush_abnormal_value: ErrorCode
    dust_full_abnormal_value: ErrorCode
    temp_high_abnormal_value: ErrorCode

    @property
    def power(self):
        if self.speed_level == SpeedOption.Medium:
            return SpeedLvPower._110W
        if self.speed_level == SpeedOption.High:
            return SpeedLvPower._400W
        return self.low_speed_lv_power

    @property
    def mif_work_duration_left(self):
        mif_max_work_duration = 36000
        if self.hepa_used_seconds < 0:
            return 0
        elif self.hepa_used_seconds == 0:
            return 1.0
        else:
            return max(
                0.0,
                1.0 - (self.hepa_used_seconds / mif_max_work_duration),
            )

    @property
    def error(self):
        if self.dust_abnormal_value != ErrorCode.OK:
            return self.dust_abnormal_value
        if self.brush_abnormal_value != ErrorCode.OK:
            return self.brush_abnormal_value
        if self.dust_full_abnormal_value != ErrorCode.OK:
            return self.dust_full_abnormal_value
        if self.temp_high_abnormal_value != ErrorCode.OK:
            return self.temp_high_abnormal_value

        if not int(self.mif_work_duration_left * 100):
            return ErrorCode.MIFDateout
        return ErrorCode.OK

    @property
    def total_clean_area(self):
        return int(self.total_clean_seconds / 60.0 * 0.8)

    def as_dict(self):
        return {
            **asdict(self),
            'state': self.state.value,
            'speed_level': self.speed_level.value,
            'actual_power': self.power.to_option(),
            'mif_duration_left': int(self.mif_work_duration_left * 100),
            'total_clean_area': self.total_clean_area,
            'error': self.error.name,
        }


class RoidmiCleaner(Sensor):
    NAME = 'roidmi_cleaner'
    VOLTAGE_CHAR = VOLTAGE_CHAR
    STATE_CHAR = STATE_CHAR
    TOTAL_USE_CHAR = TOTAL_USE_CHAR
    ERROR_CHAR = ERROR_CHAR
    ACTIVE_SLEEP_INTERVAL = 1
    RECONNECTION_SLEEP_INTERVAL = 30
    MANUFACTURER = 'Roidmi'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': STATE_ENTITY,
                    'icon': 'vacuum',
                },
                # {
                #     'name': VOLTAGE_ENTITY,
                #     'device_class': 'voltage',
                #     'unit_of_measurement': 'V',
                #     'entity_category': 'diagnostic',
                # },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
                {
                    'name': 'mif_duration_left',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'total_use_seconds',
                    'unit_of_measurement': 'S',
                    'icon': 'timer',
                },
                {
                    'name': 'total_clean_seconds',
                    'unit_of_measurement': 'S',
                    'icon': 'timer-alert',
                },
                {
                    'name': 'total_clean_area',
                    'unit_of_measurement': 'm²',
                    'icon': 'ruler-square',
                },
                {
                    'name': 'hepa_used_seconds',
                    'unit_of_measurement': 'S',
                    'icon': 'filter-variant',
                },
                {
                    'name': 'actual_power',
                    'icon': 'lightning-bolt'
                },
                {
                    'name': 'error',
                    'icon': 'alert-circle-outline',
                },
            ],
            SELECT_DOMAIN: [
                {
                    'name': LOW_SPEED_POWER,
                    'topic': LOW_SPEED_POWER,
                    'icon': 'speed',
                    'options': [x.to_option() for x in SpeedLvPower],
                },
            ],
            BUTTON_DOMAIN: [
                {
                    'name': RESET_HEPA,
                    'topic': RESET_HEPA,
                    'entity_category': 'diagnostic',
                    'device_class': 'restart',
                },
            ],
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._state = None

    @staticmethod
    def _parse_version_data(data: bytes) -> ty.Tuple[str, str]:
        mcu_version = ''
        firmware_version = ''
        if data[0] == 0xa4:
            mcu_version = data[1:7].decode('ascii').strip()
            firmware_version = data[7:13].decode('ascii').strip()
        return mcu_version, firmware_version

    @staticmethod
    def _parse_voltage_data(data: bytes) -> ty.Tuple[int, int]:
        voltage = int.from_bytes(data[1:3], byteorder='big')
        battery = 0
        if len(data) >= 4:
            battery = data[3]
        return voltage, battery

    @staticmethod
    def _parse_usage_data(data: bytes) -> ty.Tuple[int, int, int]:
        total_use_seconds = (data[1] << 16) + (data[2] << 8) + data[3]
        total_clean_seconds = (data[4] << 16) + (data[5] << 8) + data[6]
        hepa_used_seconds = (data[7] << 16) + (data[8] << 8) + data[9]
        return total_use_seconds, total_clean_seconds, hepa_used_seconds

    @staticmethod
    def _parse_state_data(data: bytes) \
            -> ty.Tuple[CleanerStateEnum, SpeedOption, SpeedLvPower, int]:
        state = CleanerStateEnum.UNKNOWN
        speed_level = SpeedOption.Low
        if data[0] == 0xa2:
            state = {
                0x51: CleanerStateEnum.WORKING,
                0x52: CleanerStateEnum.CLOSED,
                0x53: CleanerStateEnum.STANDBY,
                0x54: CleanerStateEnum.CHARGING,
                0x55: CleanerStateEnum.CHARGING,
            }.get(data[1], CleanerStateEnum.UNKNOWN)
            speed_level = {
                0x51: SpeedOption.Low,
                0x52: SpeedOption.Medium,
                0x53: SpeedOption.High,
            }.get(data[2], SpeedOption.Low)
        numeric_display_type = data[3]
        low_speed_lv_power = SpeedLvPower(
            int.from_bytes(data[4:6], byteorder='big'),
        )
        return state, speed_level, low_speed_lv_power, numeric_display_type

    @staticmethod
    def _parse_error_data(data: bytes) \
            -> ty.Tuple[ErrorCode, ErrorCode, ErrorCode, ErrorCode]:
        if data[0] == 0xa6:
            dust_abnormal_value = ErrorCode(data[1])
            brush_abnormal_value = ErrorCode(data[2])
            dust_full_abnormal_value = ErrorCode(data[3])
            temp_high_abnormal_value = ErrorCode(data[4])
            if brush_abnormal_value == ErrorCode.DustUnInstall:
                brush_abnormal_value = ErrorCode.MotorAbnormal
            return (
                dust_abnormal_value, brush_abnormal_value,
                dust_full_abnormal_value, temp_high_abnormal_value,
            )
        return ErrorCode.OK, ErrorCode.OK, ErrorCode.OK, ErrorCode.OK

    async def _read_state_from_device(self):
        voltage_bytes = await self.client.read_gatt_char(self.VOLTAGE_CHAR)
        voltage, battery = self._parse_voltage_data(voltage_bytes)
        state_bytes = await self.client.read_gatt_char(self.VOLTAGE_CHAR)
        state, speed_level, low_speed_lv_power, numeric_display_type = (
            self._parse_state_data(state_bytes))
        total_use_bytes = await self.client.read_gatt_char(self.TOTAL_USE_CHAR)
        total_use_seconds, total_clean_seconds, hepa_used_seconds = (
            self._parse_usage_data(total_use_bytes))
        error_bytes = await self.client.read_gatt_char(self.ERROR_CHAR)
        (
            dust_abnormal_value, brush_abnormal_value,
            dust_full_abnormal_value, temp_high_abnormal_value,
        ) = self._parse_error_data(error_bytes)

        return CleanerState(
            voltage=voltage,
            battery=battery,
            state=state,
            speed_level=speed_level,
            low_speed_lv_power=low_speed_lv_power,
            numeric_display_type=numeric_display_type,
            total_use_seconds=total_use_seconds,
            total_clean_seconds=total_clean_seconds,
            hepa_used_seconds=hepa_used_seconds,
            dust_abnormal_value=dust_abnormal_value,
            brush_abnormal_value=brush_abnormal_value,
            dust_full_abnormal_value=dust_full_abnormal_value,
            temp_high_abnormal_value=temp_high_abnormal_value,
        )

    def voltage_notification_handler(self, sender, data: bytearray):
        """
        sender.handle == 28
        """
        _LOGGER.debug(
            "voltage_notification_handler: {0} notification: {1}: {2}".format(
                self,
                sender,
                format_binary(data),
            )
        )
        self._state.voltage, self._state.battery = (
            self._parse_voltage_data(data))

    def state_notification_handler(self, sender, data: bytearray):
        """
        sender.handle == 32
        """
        _LOGGER.debug(
            "state_notification_handler: {0} notification: {1}: {2}".format(
                self,
                sender,
                format_binary(data),
            ),
        )
        state, speed_level, low_speed_lv_power, numeric_display_type = (
            self._parse_state_data(data))
        self._state.state = state
        self._state.speed_level = speed_level
        self._state.low_speed_lv_power = low_speed_lv_power
        self._state.numeric_display_type = numeric_display_type

    def total_use_notification_handler(self, sender, data: bytearray):
        _LOGGER.debug(
            "total_use_notification_handler: {0} notification: {1}: {2}".format(
                self,
                sender,
                format_binary(data),
            ),
        )
        total_use_seconds, total_clean_seconds, hepa_used_seconds = (
            self._parse_usage_data(data))
        self._state.total_use_seconds = total_use_seconds
        self._state.total_clean_seconds = total_clean_seconds
        self._state.hepa_used_seconds = hepa_used_seconds

    def error_notification_handler(self, sender, data: bytearray):
        _LOGGER.debug(
            "error_notification_handler: {0} notification: {1}: {2}".format(
                self,
                sender,
                format_binary(data),
            ),
        )
        (
            dust_abnormal_value, brush_abnormal_value,
            dust_full_abnormal_value, temp_high_abnormal_value,
        ) = self._parse_error_data(data)
        self._state.dust_abnormal_value = dust_abnormal_value
        self._state.brush_abnormal_value = brush_abnormal_value
        self._state.dust_full_abnormal_value = dust_full_abnormal_value
        self._state.temp_high_abnormal_value = temp_high_abnormal_value

    async def get_device_data(self):
        # don't call super(), the device doesn't have FIRMWARE_VERSION
        self._model = 'Vacuum cleaner'
        name = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(name, (bytes, bytearray)):
            self._model = name.decode().strip('\0')

        version_bytes = await self._read_with_timeout(VERSION_CHAR)
        self._version = self._parse_version_data(version_bytes)[1]
        _LOGGER.debug(f'{self} name: {self._model}, version: {self._version}')
        self._state = await self._read_state_from_device()

        _LOGGER.debug(f'{self} initial state: {self._state}')
        for ch, handler in (
            (self.VOLTAGE_CHAR, self.voltage_notification_handler),
            (self.STATE_CHAR, self.state_notification_handler),
            (self.TOTAL_USE_CHAR, self.total_use_notification_handler),
            (self.ERROR_CHAR, self.error_notification_handler),
        ):
            await self.client.start_notify(ch, handler)

    async def _set_speed_level(self, speed_level: SpeedLvPower):
        await self.client.write_gatt_char(
            self.STATE_CHAR,
            b'\xa2\x51' + speed_level.value.to_bytes(2, 'big'),
            response=False,
        )

    async def _reset_hepa_filter(self):
        await self.client.write_gatt_char(
            self.STATE_CHAR,
            b'\xa2\x52',
            response=False,
        )

    async def handle_messages(self, publish_topic, *args, **kwargs):
        while True:
            try:
                if not self.client.is_connected:
                    raise ConnectionError()
                message = await aio.wait_for(
                    self.message_queue.get(),
                    timeout=60,
                )
            except aio.TimeoutError:
                await aio.sleep(1)
                continue
            value = message['value']
            entity_topic, action_postfix = self.get_entity_subtopic_from_topic(
                message['topic'],
            )
            entity = self.get_entity_by_name(SELECT_DOMAIN, LOW_SPEED_POWER)
            if entity_topic == self._get_topic_for_entity(
                entity,
                skip_unique_id=True,
            ):
                try:
                    speed_power = SpeedLvPower.from_option(value)
                except KeyError:
                    _LOGGER.warning(f'Unknown speed option: {value}')
                    continue

                _LOGGER.info(
                    f'[{self}] switch {LOW_SPEED_POWER} value={speed_power}',
                )
                while True:
                    try:
                        await self._set_speed_level(speed_power)
                        self._state.low_speed_lv_power = speed_power
                        await aio.gather(
                            publish_topic(
                                topic=self._get_topic_for_entity(entity),
                                value=speed_power.to_option(),
                            ),
                            self._notify_state(publish_topic),
                            **get_loop_param(self._loop),
                        )
                        break
                    except ConnectionError as e:
                        _LOGGER.exception(str(e))
                    await aio.sleep(5)
                continue

            entity = self.get_entity_by_name(BUTTON_DOMAIN, RESET_HEPA)
            if entity_topic == self._get_topic_for_entity(
                entity,
                skip_unique_id=True,
            ):
                _LOGGER.info(f'[{self}] reset HEPA filter')
                while True:
                    try:
                        await self._reset_hepa_filter()
                        self._state.hepa_used_seconds = 0
                        await self._notify_state(publish_topic)
                        break
                    except ConnectionError as e:
                        _LOGGER.exception(str(e))
                    await aio.sleep(5)
                continue



import asyncio as aio
import logging
import struct
import typing as ty
import uuid
from dataclasses import dataclass

from .base import SENSOR_DOMAIN, ConnectionMode, Sensor

_LOGGER = logging.getLogger(__name__)

DEVICE_MODE_UUID = uuid.UUID('00001a00-0000-1000-8000-00805f9b34fb')
DATA_UUID = uuid.UUID('00001a01-0000-1000-8000-00805f9b34fb')
FIRMWARE_UUID = uuid.UUID('00001a02-0000-1000-8000-00805f9b34fb')

LIVE_MODE_CMD = bytes([0xA0, 0x1F])


@dataclass
class SensorState:
    temperature: float
    moisture: int
    illuminance: ty.Optional[int]
    conductivity: int
    battery: int = 0

    @classmethod
    def from_data(cls, data: bytes, battery: int):
        if len(data) == 24:  # is ropot
            light = None
            temp, moist, conductivity = struct.unpack(
                "<hxxxxxBhxxxxxxxxxxxxxx", data
            )
        else:
            temp, light, moist, conductivity = struct.unpack(
                "<hxIBhxxxxxx", data
            )

        return cls(
            temperature=temp / 10.0,
            moisture=moist,
            conductivity=conductivity,
            illuminance=light,
            battery=battery,
        )


class FlowerMonitorMiFlora(Sensor):
    NAME = 'miflora'
    MANUFACTURER = 'Xiaomi'
    MODEL = 'MiFlora'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT
    DEFAULT_RECONNECTION_SLEEP_INTERVAL = 300
    READ_DATA_IN_ACTIVE_LOOP = True

    def __init__(self, mac, *args, **kwargs) -> None:
        super().__init__(mac, *args, **kwargs)
        self.RECONNECTION_SLEEP_INTERVAL = int(
            kwargs.get('interval', self.DEFAULT_RECONNECTION_SLEEP_INTERVAL)
        )

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'moisture',
                    'device_class': 'moisture',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'illuminance',
                    'device_class': 'illuminance',
                    'unit_of_measurement': 'lx',
                },
                {
                    'name': 'conductivity',
                    'unit_of_measurement': 'µS/cm',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }

    async def read_state(self):
        fw_and_bat = await self._read_with_timeout(FIRMWARE_UUID)
        battery = fw_and_bat[0]
        data = await self._read_with_timeout(DATA_UUID)
        for _ in range(5):
            try:
                self._state = SensorState.from_data(data, battery)
            except ValueError as e:
                _LOGGER.warning(f'{self} {repr(e)}')
                await aio.sleep(1)
            else:
                break

    async def get_device_data(self):
        self._model = self.MODEL
        await self.client.write_gatt_char(
            DEVICE_MODE_UUID,
            LIVE_MODE_CMD,
            response=False,
        )
        fw_and_bat = await self._read_with_timeout(FIRMWARE_UUID)
        self._version = fw_and_bat[2:].decode('ascii').strip('\0')

    async def do_active_loop(self, publish_topic):
        try:
            await aio.wait_for(self.read_state(), 5)
            await self._notify_state(publish_topic)
        except (aio.CancelledError, KeyboardInterrupt):
            raise
        except Exception:
            _LOGGER.exception(f'{self} problem with reading values')



from dataclasses import dataclass

from ..devices.base import (SENSOR_DOMAIN, ConnectionMode, Sensor,
                            SubscribeAndSetDataMixin)

UUID_KEY_READ = "0000fff4-0000-1000-8000-00805f9b34fb"
KEY = b"\x6c\x65\x61\x67\x65\x6e\x64\xff\xfe\x31\x38\x38\x32\x34\x36\x36"


def create_aes():
    try:
        from Crypto.Cipher import AES
    except ImportError:
        raise ImportError(
            "Please install pycryptodome to setup BM2 Voltage meter",
        ) from None

    return AES.new(KEY, AES.MODE_CBC, bytes([0] * 16))


@dataclass
class SensorState:
    voltage: float

    @classmethod
    def from_data(cls, decrypted_data: bytes):
        voltage = (int.from_bytes(
            decrypted_data[1:1 + 2],
            byteorder='big',
        ) >> 4) / 100
        return cls(voltage=round(voltage, 2))


class VoltageTesterBM2(SubscribeAndSetDataMixin, Sensor):
    NAME = 'voltage_bm2'
    DATA_CHAR = UUID_KEY_READ
    MANUFACTURER = 'BM2'
    SENSOR_CLASS = SensorState
    REQUIRED_VALUES = ('voltage', )
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'voltage',
                    'device_class': 'voltage',
                    'unit_of_measurement': 'V',
                },
            ],
        }

    def process_data(self, data: bytearray):
        decrypted_data = create_aes().decrypt(data)
        if decrypted_data[0] == 0xf5:
            self._state = self.SENSOR_CLASS.from_data(decrypted_data)



import asyncio as aio
import logging
import os
from contextlib import asynccontextmanager

import aio_mqtt
from bleak import BleakError

ListOfConnectionErrors = (
    BleakError,
    aio.TimeoutError,

    # dbus-next exceptions:
    # AttributeError: 'NoneType' object has no attribute 'call'
    AttributeError,
    # https://github.com/hbldh/bleak/issues/409
    EOFError,
)

ListOfMQTTConnectionErrors = (
        aio_mqtt.ConnectionLostError,
        aio_mqtt.ConnectionClosedError,
        aio_mqtt.ServerDiedError,
        BrokenPipeError,
)


_LOGGER = logging.getLogger(__name__)

# initialize in a loop
BLUETOOTH_RESTARTING = aio.Lock()


def hardware_exception_occurred(exception):
    ex_str = str(exception)
    return (
        'org.freedesktop.DBus.Error.ServiceUnknown' in ex_str or
        'org.freedesktop.DBus.Error.NoReply' in ex_str or
        'org.freedesktop.DBus.Error.AccessDenied' in ex_str or
        'org.bluez.Error.Failed: Connection aborted' in ex_str or
        'org.bluez.Error.NotReady' in ex_str or
        'org.bluez.Error.InProgress' in ex_str
    )


async def restart_bluetooth(adapter: str):
    if BLUETOOTH_RESTARTING.locked():
        await aio.sleep(9)
        return
    async with BLUETOOTH_RESTARTING:
        _LOGGER.warning('Restarting bluetoothd...')
        proc = await aio.create_subprocess_exec(
            'hciconfig', adapter, 'down',
        )
        await proc.wait()

        if os.path.exists('/etc/init.d/bluetoothd'):
            proc = await aio.create_subprocess_exec(
                '/etc/init.d/bluetoothd', 'restart',
            )
            await proc.wait()

        elif os.path.exists('/etc/init.d/bluetooth'):
            proc = await aio.create_subprocess_exec(
                '/etc/init.d/bluetooth', 'restart',
            )
            await proc.wait()

        else:
            _LOGGER.error('init.d bluetoothd script not found')

        await aio.sleep(3)
        proc = await aio.create_subprocess_exec(
            'hciconfig', adapter, 'up',
        )
        await proc.wait()
        await aio.sleep(5)
        _LOGGER.warning('Restarting bluetoothd finished')


@asynccontextmanager
async def handle_ble_exceptions(adapter: str):
    try:
        yield
    except ListOfConnectionErrors as e:
        if hardware_exception_occurred(e):
            await restart_bluetooth(adapter)
            await aio.sleep(3)
        raise



import logging
import uuid
from dataclasses import dataclass

from .base import ConnectionMode
from .uuids import BATTERY
from .xiaomi_base import XiaomiHumidityTemperature

_LOGGER = logging.getLogger(__name__)

SERVICE_DATA_UUID = uuid.UUID('0000fdcd-0000-1000-8000-00805f9b34fb')


@dataclass
class SensorState:
    battery: int = 0
    temperature: float = 0.0
    humidity: float = 0.0


class QingpingTempRHMonitorLite(XiaomiHumidityTemperature):
    NAME = 'qingpingCGDK2'
    MANUFACTURER = 'Qingping'
    DATA_CHAR = SERVICE_DATA_UUID
    BATTERY_CHAR = BATTERY
    SENSOR_CLASS = SensorState
    SUPPORT_PASSIVE = True
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT

    PREAMBLE = b'\xcd\xfd'

    def filter_notifications(self, sender, data):
        packet_start = data.find(self.PREAMBLE)
        if packet_start == -1:
            return False
        return data[packet_start + len(self.PREAMBLE) + 1] == 0x10

    def process_data(self, data):
        packet_start = data.find(self.PREAMBLE)
        offset = packet_start + len(self.PREAMBLE)
        value_data = data[offset:]

        self._state = self.SENSOR_CLASS(
            temperature=int.from_bytes(
                value_data[10:12],
                byteorder='little',
                signed=True,
            ) / 10,
            humidity=int.from_bytes(
                value_data[12:14],
                byteorder='little',
                signed=False,
            ) / 10,
            battery=value_data[16],
        )



import asyncio as aio
import json
import logging
import typing as ty

import aio_mqtt
from bleak.backends.device import BLEDevice

from ble2mqtt.__version__ import VERSION

from .devices.base import (BINARY_SENSOR_DOMAIN, BUTTON_DOMAIN, CLIMATE_DOMAIN,
                           COVER_DOMAIN, DEVICE_TRACKER_DOMAIN, LIGHT_DOMAIN,
                           SELECT_DOMAIN, SENSOR_DOMAIN, SWITCH_DOMAIN,
                           ConnectionMode, ConnectionTimeoutError, Device)
from .exceptions import (BLUETOOTH_RESTARTING, ListOfConnectionErrors,
                         ListOfMQTTConnectionErrors, handle_ble_exceptions,
                         restart_bluetooth)
from .tasks import handle_returned_tasks, run_tasks_and_cancel_on_first_return

_LOGGER = logging.getLogger(__name__)

CONFIG_MQTT_NAMESPACE = 'homeassistant'
FAILURE_LIMIT = 5


class DeviceManager:
    def __init__(self, device, *, hci_adapter, mqtt_client, base_topic,
                 config_prefix, global_availability_topic, legacy_color_mode):
        self.device: Device = device
        self._hci_adapter = hci_adapter
        self._mqtt_client = mqtt_client
        self._base_topic = base_topic
        self._config_prefix = config_prefix
        self._global_availability_topic = global_availability_topic
        self._legacy_color_mode = legacy_color_mode
        self.manage_task = None
        self.last_connection_successful = True

        self._scanned_device: ty.Union[BLEDevice, None] = None
        self._scanned_device_set = aio.Event()

    def set_scanned_device(self, ble_device: BLEDevice):
        self._scanned_device = ble_device
        self._scanned_device_set.set()

    async def close(self):
        if self.manage_task and not self.manage_task.done():
            self.manage_task.cancel()
            try:
                await self.manage_task
            except aio.CancelledError:
                pass
        self.manage_task = None
        try:
            await self.device.close()
        except aio.CancelledError:
            raise
        except Exception:
            _LOGGER.exception(f'Problem on closing device {self.device}')

    def run_task(self) -> aio.Task:
        assert not self.manage_task, \
            f'{self.device} Previous task was not finished! {self.manage_task}'
        self.manage_task = aio.create_task(self.manage_device())
        return self.manage_task

    async def publish_topic_callback(self, topic, value, nowait=False):
        _LOGGER.debug(f'call publish callback topic={topic} value={value}')
        if not self._mqtt_client.is_connected():
            _LOGGER.warning(f'{self.device} mqtt is disconnected')
            return
        try:
            await self._mqtt_client.publish(
                aio_mqtt.PublishableMessage(
                    topic_name='/'.join((self._base_topic, topic)),
                    payload=value,
                    qos=aio_mqtt.QOSLevel.QOS_1,
                ),
                nowait=nowait,
            )
        except ListOfMQTTConnectionErrors:
            _LOGGER.exception('Error while publishing to MQTT')

    def _get_topic(self, dev_id, subtopic, *args):
        return '/'.join(
            filter(None, (self._base_topic, dev_id, subtopic, *args)),
        )

    @property
    def _config_device_topic(self):
        """Add a prefix to avoid interfering with other ble software"""

        return f'{self._config_prefix}{self.device.dev_id}'

    async def send_device_config(self):
        device = self.device
        device_info = {
            'identifiers': [
                device.unique_id,
            ],
            'name': device.unique_name,
            'model': device.model,
        }
        if device.manufacturer:
            device_info['manufacturer'] = device.manufacturer
        if device.version:
            device_info['sw_version'] = device.version
        if device.suggested_area:
            device_info['suggested_area'] = device.suggested_area

        def get_generic_vals(entity: dict):
            name = entity.pop('name')
            result = {
                'name': f'{name}_{device.friendly_id}',
                'unique_id': f'{name}_{device.dev_id}',
                'device': device_info,
                'availability_mode': 'all',
                'availability': [
                    {'topic': self._global_availability_topic},
                    {'topic': '/'.join(
                        (self._base_topic, self.device.availability_topic),
                    )},
                ],
                'origin': {'name': 'ble2mqtt', 'sw_version': VERSION},
            }
            icon = entity.pop('icon', None)
            if icon:
                result['icon'] = f'mdi:{icon}'
            entity.pop('topic', None)
            entity.pop('json', None)
            entity.pop('main_value', None)
            result.update(entity)
            return result

        messages_to_send = []
        for cls, entities in device.entities_with_lqi.items():
            if cls in (
                BINARY_SENSOR_DOMAIN,
                SENSOR_DOMAIN,
                DEVICE_TRACKER_DOMAIN,
            ):
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    if entity.get('json') and entity.get('main_value'):
                        state_topic_part = {
                            'json_attributes_topic': state_topic,
                            'state_topic': state_topic,
                            'value_template':
                                f'{{{{ value_json.{entity["main_value"]} }}}}',
                        }
                    else:
                        state_topic_part = {
                            'state_topic': state_topic,
                            'value_template':
                                f'{{{{ value_json.{entity_name} }}}}',
                        }

                    if cls == DEVICE_TRACKER_DOMAIN:
                        state_topic_part['source_type'] = 'bluetooth_le'

                    payload = json.dumps({
                        **get_generic_vals(entity),
                        **state_topic_part,
                    })
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls in {BUTTON_DOMAIN, SWITCH_DOMAIN}:
                has_state = cls == SWITCH_DOMAIN
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    command_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    payload = json.dumps({
                        **get_generic_vals(entity),
                        **({'state_topic': state_topic} if has_state else {}),
                        'command_topic': command_topic,
                    })
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
                    if has_state:
                        # TODO: send real state on receiving status
                        # from a device
                        _LOGGER.debug(
                            f'Publish initial state topic={state_topic}')
                        await self._mqtt_client.publish(
                            aio_mqtt.PublishableMessage(
                                topic_name=state_topic,
                                payload='OFF',
                                qos=aio_mqtt.QOSLevel.QOS_1,
                            ),
                        )
            if cls == LIGHT_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    set_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    color_mode = entity['color_mode']
                    payload = json.dumps({
                        **get_generic_vals(entity),
                        'schema': 'json',
                        **(
                            {'color_mode': bool(color_mode)}
                            if self._legacy_color_mode
                            else {}
                        ),
                        'supported_color_modes': entity.get(
                            'supported_color_modes', [color_mode],
                        ),
                        'brightness': entity.get('brightness', True),
                        'state_topic': state_topic,
                        'command_topic': set_topic,
                    })
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == COVER_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    set_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    set_position_topic = '/'.join(
                        (state_topic, device.SET_POSITION_POSTFIX),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    config_params = {
                        **get_generic_vals(entity),
                        'state_topic': state_topic,
                        'position_topic': state_topic,
                        'json_attributes_topic': state_topic,
                        'value_template': '{{ value_json.state }}',
                        'position_template': '{{ value_json.position }}',
                        'command_topic': set_topic,
                        'set_position_topic': set_position_topic,
                    }
                    payload = json.dumps(config_params)
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == SELECT_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    set_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    config_params = {
                        **get_generic_vals(entity),
                        'state_topic': state_topic,
                        'command_topic': set_topic,
                    }
                    payload = json.dumps(config_params)
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == CLIMATE_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    mode_command_topic = '/'.join(
                        (state_topic, device.SET_MODE_POSTFIX),
                    )
                    temperature_command_topic = '/'.join(
                        (state_topic, device.SET_TARGET_TEMPERATURE_POSTFIX),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    config_params = {
                        **get_generic_vals(entity),
                        'current_temperature_topic': state_topic,
                        'current_temperature_template':
                            '{{ value_json.temperature }}',
                        'mode_state_topic': state_topic,
                        'mode_state_template': '{{ value_json.mode }}',
                        'mode_command_topic': mode_command_topic,
                        'temperature_state_topic': state_topic,
                        'temperature_state_template':
                            '{{ value_json.target_temperature }}',
                        'temperature_command_topic': temperature_command_topic,
                        'json_attributes_topic': state_topic,
                        'temp_step': 0.5,
                    }
                    payload = json.dumps(config_params)
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
        await aio.gather(*[
            self._mqtt_client.publish(message)
            for message in messages_to_send
        ])
        device.config_sent = True

    async def send_availability(self, value: bool):
        return await self.device.send_availability(
            self.publish_topic_callback,
            value,
        )

    async def _sleep_until_next_connection(self):
        device = self.device
        _LOGGER.debug(
            f'Sleep for {device.RECONNECTION_SLEEP_INTERVAL} secs to '
            f'reconnect to device={device}',
        )
        if (
            device._connection_mode == ConnectionMode.ACTIVE_KEEP_CONNECTION or
            # if last connection failed, connect as soon as device appears,
            # don't wait for RECONNECTION_SLEEP_INTERVAL seconds
            not self.last_connection_successful
        ):
            try:
                await aio.wait_for(
                    device._advertisement_seen.wait(),
                    timeout=device.RECONNECTION_SLEEP_INTERVAL,
                )
            except aio.TimeoutError:
                pass
        else:
            await aio.sleep(self.device.RECONNECTION_SLEEP_INTERVAL)

    async def publish_topic_with_availability(self, topic, value):
        # call sequentially to allow HA receive a new value
        await self.publish_topic_callback(topic, value)
        await self.send_availability(True)

    async def manage_device(self):
        device = self.device
        _LOGGER.debug(f'Start managing device={device}')
        failure_count = 0
        missing_device_count = 0
        while True:
            async with BLUETOOTH_RESTARTING:
                _LOGGER.debug(f'[{device}] Check for lock')
            try:
                self.last_connection_successful = False
                if not device.is_passive:
                    try:
                        await aio.wait_for(
                            self._scanned_device_set.wait(),
                            timeout=10,
                        )
                    except aio.TimeoutError as e:
                        raise ConnectionTimeoutError(
                            f'[{device}] is not visible for 10 sec',
                        ) from e
                async with handle_ble_exceptions(self._hci_adapter):
                    await device.connect(
                        self._hci_adapter,
                        self._scanned_device,
                    )
                    self.last_connection_successful = True
                    initial_coros = []
                    if not device.is_passive:
                        if not device.DEVICE_DROPS_CONNECTION:
                            initial_coros.append(device.disconnected_event.wait)
                        await device.get_device_data()
                        failure_count = 0
                        missing_device_count = 0

                    if device.subscribed_topics:
                        await self._mqtt_client.subscribe(*[
                            (
                                '/'.join((self._base_topic, topic)),
                                aio_mqtt.QOSLevel.QOS_1,
                            )
                            for topic in device.subscribed_topics
                        ])
                    _LOGGER.debug(f'[{device}] mqtt subscribed')
                    coros = [
                        *[coro() for coro in initial_coros],
                        device.handle(
                            self.publish_topic_with_availability,
                            send_config=self.send_device_config,
                        ),
                    ]
                    will_handle_messages = bool(device.subscribed_topics)
                    if will_handle_messages:
                        coros.append(
                            device.handle_messages(
                                self.publish_topic_with_availability,
                            ),
                        )

                    tasks = [aio.create_task(t) for t in coros]
                    _LOGGER.debug(f'[{device}] tasks are created')

                    await run_tasks_and_cancel_on_first_return(*tasks)
                    if device.disconnected_event.is_set():
                        _LOGGER.debug(f'{device} has disconnected')
                    finished = [t for t in tasks if not t.cancelled()]
                    await handle_returned_tasks(*finished)
            except aio.CancelledError:
                # the only way to send availability=False on program shutdown
                if self.device.ACTIVE_CONNECTION_MODE in (
                    ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT,
                    ConnectionMode.ON_DEMAND_CONNECTION,
                ):
                    try:
                        await aio.wait_for(
                            self.send_availability(False),
                            timeout=1,
                        )
                    except aio.TimeoutError:
                        pass
                raise
            except KeyboardInterrupt:
                raise
            except ConnectionTimeoutError:
                missing_device_count += 1
                _LOGGER.error(
                    f'[{device}] connection problem, '
                    f'attempts={missing_device_count}',
                )
            except (ConnectionError, TimeoutError, aio.TimeoutError):
                missing_device_count += 1
                _LOGGER.exception(
                    f'[{device}] connection problem, '
                    f'attempts={missing_device_count}',
                )
            except ListOfConnectionErrors as e:
                if 'Device with address' in str(e) and \
                        'was not found' in str(e):
                    missing_device_count += 1
                    _LOGGER.warning(
                        f'Error while connecting to {device}, {e} {repr(e)}, '
                        f'attempts={missing_device_count}',
                    )
                else:
                    # if isinstance(e, aio.TimeoutError) or \
                    #         'org.bluez.Error.Failed: Connection aborted' in \
                    #         str(e):
                    failure_count += 1
                    _LOGGER.warning(
                        f'Error while connecting to {device}, {e} {repr(e)}, '
                        f'failure_count={failure_count}',
                    )

                # sometimes LYWSD03MMC devices remain connected
                # and doesn't advert their presence.
                # If cannot find device for several attempts, restart
                # the bluetooth chip
                if missing_device_count >= device.CONNECTION_FAILURES_LIMIT:
                    _LOGGER.error(
                        f'Device {device} was not found for '
                        f'{missing_device_count} times. Restarting bluetooth.',
                    )
                    missing_device_count = 0
                    await restart_bluetooth(self._hci_adapter)
            finally:
                if self.device.ACTIVE_CONNECTION_MODE not in (
                    ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT,
                    ConnectionMode.ON_DEMAND_CONNECTION,
                ):
                    try:
                        await aio.wait_for(
                            self.send_availability(False),
                            timeout=1,
                        )
                    except aio.TimeoutError:
                        pass
                try:
                    await aio.wait_for(device.close(), timeout=5)
                except aio.CancelledError:
                    raise
                except Exception:
                    _LOGGER.exception(f'{device} problem on device.close()')
                try:
                    canceled = []
                    for t in tasks:
                        if not t.done():
                            t.cancel()
                            canceled.append(t)
                    for t in canceled:
                        try:
                            t.result()
                        except aio.CancelledError:
                            pass
                except aio.CancelledError:
                    raise
                except Exception:
                    pass

            if failure_count >= FAILURE_LIMIT:
                await restart_bluetooth(self._hci_adapter)
                failure_count = 0
            try:
                if not device.disconnected_event.is_set():
                    await aio.wait_for(
                        device.disconnected_event.wait(),
                        timeout=10,
                    )
            except aio.TimeoutError:
                _LOGGER.exception(f'{device} not disconnected in 10 secs')
            await self._sleep_until_next_connection()



import asyncio as aio
import logging
import typing as ty

_LOGGER = logging.getLogger(__name__)


async def run_tasks_and_cancel_on_first_return(*tasks: aio.Future,
                                               return_when=aio.FIRST_COMPLETED,
                                               ignore_futures=(),
                                               ) -> ty.Sequence[aio.Future]:
    async def cancel_tasks(_tasks) -> ty.List[aio.Task]:
        # cancel first, then await. Because other tasks can raise exceptions
        # while switching tasks
        canceled = []
        for t in _tasks:
            if t in ignore_futures:
                continue
            if not t.done():
                t.cancel()
                canceled.append(t)
        tasks_raise_exceptions = []
        for t in canceled:
            try:
                await t
            except aio.CancelledError:
                pass
            except Exception:
                _LOGGER.exception(
                    f'Unexpected exception while cancelling tasks! {t}',
                )
                tasks_raise_exceptions.append(t)
        return tasks_raise_exceptions

    assert all(isinstance(t, aio.Future) for t in tasks)
    try:
        # NB: pending tasks can still raise exception or finish
        # while tasks are switching
        done, pending = await aio.wait(tasks, return_when=return_when)
    except aio.CancelledError:
        await cancel_tasks(tasks)
        # it could happen that tasks raised exception and canceling wait task
        # abandons tasks with exception
        for t in tasks:
            if not t.done() or t.cancelled():
                continue
            try:
                t.result()
            # no CancelledError expected
            except Exception:
                _LOGGER.exception(
                    f'Task raises exception while cancelling parent coroutine '
                    f'that waits for it {t}')
        raise

    # while switching tasks for await other pending tasks can raise an exception
    # we need to append more tasks to the result if so
    await cancel_tasks(pending)

    task_remains = [t for t in pending if not t.cancelled()]
    return [*done, *task_remains]


async def handle_returned_tasks(*tasks: aio.Future):
    raised = [t for t in tasks if t.done() and t.exception()]
    returned_normally = set(tasks) - set(raised)

    results = []

    if raised:
        task_for_raise = raised.pop()
        for t in raised:
            try:
                await t
            except aio.CancelledError:
                raise
            except Exception:
                _LOGGER.exception('Task raised an error')
        await task_for_raise
    for t in returned_normally:
        results.append(await t)
    return results
